@@ -4,6 +4,7 @@
 #include "database.h"
 #include "question.h"
 #include "template.h"
+#include "plugin.h"
 #include "strutl.h"
 
 #include <dlfcn.h>
@@ -112,17 +113,44 @@ command_capb(struct confmodule *mod, char *arg)
     int i;
     char *argv[32];
     int argc;
-    char *out;
+    char *out, *outend;
+    size_t outalloc;
+    struct plugin *plugin;
+    void *plugin_state;
 
     argc = strcmdsplit(arg, argv, DIM(argv));
     /* FIXME: frontend.h should provide a method to prevent direct
      *        access to capability member */
     mod->frontend->capability = 0;
-    for (i = 0; i < argc; i++)
+    for (i = 0; i < argc; i++) {
         if (strcmp(argv[i], "backup") == 0)
             mod->frontend->capability |= DCF_CAPB_BACKUP;
+        else if (strcmp(argv[i], "progresscancel") == 0)
+            mod->frontend->capability |= DCF_CAPB_PROGRESSCANCEL;
+    }
+
+    if (asprintf(&out, "%u multiselect backup progresscancel", CMDSTATUS_SUCCESS) == -1)
+        DIE("Out of memory");
+
+    plugin_state = NULL;
+    outend = strchr(out, '\0');
+    outalloc = outend - out + 1;
+    while ((plugin = plugin_iterate(mod->frontend, &plugin_state)) != NULL) {
+        size_t namelen;
+        char *newout;
+
+        namelen = strlen(plugin->name);
+        outalloc += 8 + namelen;
+        newout = realloc(out, outalloc);
+        if (!newout)
+            DIE("Out of memory");
+        outend = newout + (outend - out);
+        out = newout;
+        outend = mempcpy(outend, " plugin-", 8);
+        outend = mempcpy(outend, plugin->name, namelen);
+        *outend++ = '\0';
+    }
 
-    asprintf(&out, "%u multiselect backup", CMDSTATUS_SUCCESS);
     return out;
 }
 
@@ -131,7 +159,8 @@ command_title(struct confmodule *mod, char *arg)
 {
     char *out;
 
-    debug_printf(0, "Obsolete command TITLE %s called", arg);
+    if (*arg)
+        debug_printf(0, "Obsolete command TITLE %s called", arg);
 
     mod->frontend->methods.set_title(mod->frontend, arg);
     asprintf(&out, "%u", CMDSTATUS_SUCCESS);
@@ -162,16 +191,17 @@ command_go(struct confmodule *mod, char *arg)
     char *argv[3];
     int argc;
     char *out;
-    char *running_frontend = NULL;
-    char *requested_frontend = NULL;
+    const char *running_frontend = NULL;
+    const char *requested_frontend = NULL;
     struct question *q;
+    int ret;
 
     argc = strcmdsplit(arg, argv, DIM(argv) - 1);
     CHECKARGC(== 0);
 
     q = mod->questions->methods.get(mod->questions, "debconf/frontend");
     if (q)
-	requested_frontend = question_getvalue(q, "");
+	requested_frontend = question_getvalue(q, "C");
     question_deref(q);
 
     running_frontend = getenv("DEBIAN_FRONTEND");
@@ -187,12 +217,19 @@ command_go(struct confmodule *mod, char *arg)
 	mod->frontend->questions = q;
     }
 
-    if (mod->frontend->methods.go(mod->frontend) == CMDSTATUS_GOBACK || mod->backed_up != 0)
+    ret = mod->frontend->methods.go(mod->frontend);
+    if (ret == CMDSTATUS_GOBACK || mod->backed_up != 0)
     {
         mod->backed_up = 1;
         asprintf(&out, "%u backup", CMDSTATUS_GOBACK);
         mod->update_seen_questions(mod, STACK_SEEN_REMOVE);
     }
+    else if (ret == DC_NOTOK) /* TODO return value namespace */
+    {
+        mod->backed_up = 0;
+        asprintf(&out, "%u internal error", CMDSTATUS_INTERNALERROR);
+        mod->update_seen_questions(mod, STACK_SEEN_REMOVE);
+    }
     else
     {
         mod->backed_up = 0;
@@ -221,7 +258,7 @@ command_get(struct confmodule *mod, char *arg)
                 CMDSTATUS_BADQUESTION, argv[0]);
     else 
     {
-        const char *value = question_getvalue(q, NULL);
+        const char *value = question_getvalue(q, "C");
         asprintf(&out, "%u %s",
                 CMDSTATUS_SUCCESS, value ? value : "");
     }
@@ -247,6 +284,7 @@ command_set(struct confmodule *mod, char *arg)
                 CMDSTATUS_BADQUESTION, argv[0]);
     else
     {
+        char *prev = STRDUP(question_getvalue(q, ""));
         question_setvalue(q, argv[1]);
 
         if (mod->questions->methods.set(mod->questions, q) != 0)
@@ -256,6 +294,13 @@ command_set(struct confmodule *mod, char *arg)
             { /* Pass the value on to getlanguage() in templates.c */
                 debug_printf(0, "Setting debconf/language to %s", argv[1]);
                 setenv("LANGUAGE", argv[1], 1);
+                /* Reloading takes a little while, so only do it when it's
+                 * really necessary.
+                 */
+                if (!load_all_translations() &&
+                    strcmp(argv[1], "C") != 0 && strcmp(argv[1], "en") != 0 &&
+                    (prev == NULL || strcmp(argv[1], prev) != 0))
+                    mod->templates->methods.reload(mod->templates);
             }
             else if (strcmp(argv[0], "debconf/priority") == 0)
             {
@@ -265,6 +310,8 @@ command_set(struct confmodule *mod, char *arg)
         }
         else
             asprintf(&out, "%u cannot set value", CMDSTATUS_INTERNALERROR);
+
+        free(prev);
     }
     question_deref(q);
 
@@ -358,8 +405,9 @@ command_register(struct confmodule *mod, char *arg)
     }
     question_owner_add(q, mod->owner);
     q->template = t;
-    template_ref(t);
+        /* steal reference from mod->templates->methods.get above */
     mod->questions->methods.set(mod->questions, q);
+    question_deref(q);
     asprintf(&out, "%u", CMDSTATUS_SUCCESS);
     return out;
 }
@@ -381,6 +429,7 @@ command_unregister(struct confmodule *mod, char *arg)
         return out;
     }
     question_owner_delete(q, mod->owner);
+    question_deref(q);
     asprintf(&out, "%u", CMDSTATUS_SUCCESS);
 
     return out;
@@ -400,7 +449,7 @@ char *
 command_metaget(struct confmodule *mod, char *arg)
 {
     struct question *q;
-    const char *value;
+    char *value;
     char *argv[4];
     int argc;
     char *out;
@@ -419,6 +468,9 @@ command_metaget(struct confmodule *mod, char *arg)
         asprintf(&out, "%u %s does not exist", CMDSTATUS_BADQUESTION, argv[1]);
     else
         asprintf(&out, "%u %s", CMDSTATUS_SUCCESS, value);
+    free(value);
+
+    question_deref(q);
 
     return out;
 }
@@ -533,12 +585,13 @@ command_x_loadtemplatefile(struct confmodule *mod, char *arg)
 {
     struct template *t = NULL;
     struct question *q = NULL;
-    char *argv[3];
+    char *argv[3] = { "", "", "" };
     int argc;
     char *out;
 
     argc = strcmdsplit(arg, argv, DIM(argv));
-    CHECKARGC(== 1);
+    CHECKARGC(>= 1);
+    CHECKARGC(<= 2);
     t = template_load(argv[0]);
     while (t)
     {
@@ -549,7 +602,15 @@ command_x_loadtemplatefile(struct confmodule *mod, char *arg)
             q->template = t;
             template_ref(t);
         }
+        else if (q->template != t) {
+            template_deref(q->template);
+            q->template = t;
+            template_ref(t);
+        }
+        if (*argv[1])
+            question_owner_add(q, argv[1]);
         mod->questions->methods.set(mod->questions, q);
+        question_deref(q);
         t = t->next;
     }
     asprintf(&out, "%u OK", CMDSTATUS_SUCCESS);
@@ -561,10 +622,11 @@ command_progress(struct confmodule *mod, char *arg)
 {
     int min, max;
     struct question *q = NULL;
-    const char *value;
+    char *value;
     char *argv[6];
     int argc;
     char *out;
+    int ret;
 
     argc = strcmdsplit(arg, argv, DIM(argv));
     CHECKARGC(>= 1);
@@ -591,6 +653,7 @@ command_progress(struct confmodule *mod, char *arg)
             return out;
         }
         value = question_get_field(q, "", "description");
+        question_deref(q);
         if (value == NULL)
         {
             asprintf(&out, "%u %s description field does not exist",
@@ -599,16 +662,29 @@ command_progress(struct confmodule *mod, char *arg)
         }
         mod->frontend->methods.progress_start(mod->frontend,
                 min, max, value);
+        free(value);
     }
     else if (strcasecmp(argv[0], "set") == 0)
     {
         CHECKARGC(== 2);
-        mod->frontend->methods.progress_set(mod->frontend, atoi(argv[1]));
+        ret = mod->frontend->methods.progress_set(mod->frontend, atoi(argv[1]));
+        if (ret == DC_GOBACK)
+        {
+            asprintf(&out, "%u progress bar cancelled",
+                    CMDSTATUS_PROGRESSCANCELLED);
+            return out;
+        }
     }
     else if (strcasecmp(argv[0], "step") == 0)
     {
         CHECKARGC(== 2);
-        mod->frontend->methods.progress_step(mod->frontend, atoi(argv[1]));
+        ret = mod->frontend->methods.progress_step(mod->frontend, atoi(argv[1]));
+        if (ret == DC_GOBACK)
+        {
+            asprintf(&out, "%u progress bar cancelled",
+                    CMDSTATUS_PROGRESSCANCELLED);
+            return out;
+        }
     }
     else if (strcasecmp(argv[0], "info") == 0)
     {
@@ -621,13 +697,21 @@ command_progress(struct confmodule *mod, char *arg)
             return out;
         }
         value = question_get_field(q, "", "description");
+        question_deref(q);
         if (value == NULL)
         {
             asprintf(&out, "%u %s description field does not exist",
                     CMDSTATUS_BADQUESTION, argv[1]);
             return out;
         }
-        mod->frontend->methods.progress_info(mod->frontend, value);
+        ret = mod->frontend->methods.progress_info(mod->frontend, value);
+        free(value);
+        if (ret == DC_GOBACK)
+        {
+            asprintf(&out, "%u progress bar cancelled",
+                    CMDSTATUS_PROGRESSCANCELLED);
+            return out;
+        }
     }
     else if (strcasecmp(argv[0], "stop") == 0)
     {
@@ -646,7 +730,7 @@ char *
 command_settitle(struct confmodule *mod, char *arg)
 {
     struct question *q = NULL;
-    const char *value;
+    char *value;
     char *out;
 
     q = mod->questions->methods.get(mod->questions, arg);
@@ -656,6 +740,7 @@ command_settitle(struct confmodule *mod, char *arg)
 	return out;
     }
     value = question_get_field(q, "", "description");
+    question_deref(q);
     if (value == NULL)
     {
 	asprintf(&out, "%u %s description field does not exist",
@@ -664,7 +749,99 @@ command_settitle(struct confmodule *mod, char *arg)
     }
 
     mod->frontend->methods.set_title(mod->frontend, value);
+    free(value);
 
     asprintf(&out, "%u OK", CMDSTATUS_SUCCESS);
     return out;
 }
+
+char *command_x_save(struct confmodule *mod, char *arg)
+{
+    char *argv[2];
+    int argc, ret = DC_OK;
+    char *out;
+
+    argc = strcmdsplit(arg, argv, DIM(argv));
+    CHECKARGC(== 0);
+    if (mod)
+        ret = mod->save(mod);
+    if (ret == DC_OK)
+        asprintf(&out, "%u OK", CMDSTATUS_SUCCESS);
+    else
+        asprintf(&out, "%u not OK", CMDSTATUS_INTERNALERROR);
+    return out;
+}
+
+/*
+ * Displays the given template as an out-of-band informative message. Unlike
+ * inputting a note, this doesn't require an acknowledgement from the user,
+ * and depending on the frontend it may not even be displayed at all. Like
+ * PROGRESS INFO, frontends should display the info persistently until some
+ * other info comes along.
+ */
+char *
+command_info(struct confmodule *mod, char *arg)
+{
+    struct question *q = NULL;
+    char *out;
+
+    q = mod->questions->methods.get(mod->questions, arg);
+    if (q == NULL)
+    {
+        asprintf(&out, "%u %s does not exist", CMDSTATUS_BADQUESTION, arg);
+        return out;
+    }
+    mod->frontend->methods.info(mod->frontend, q);
+    question_deref(q);
+
+    asprintf(&out, "%u OK", CMDSTATUS_SUCCESS);
+    return out;
+}
+
+/*
+ * Accept template data from the client
+ * (for use with the passthrough frontend)
+ */
+char *
+command_data(struct confmodule *mod, char *arg)
+{
+    char *argv[3];
+    const char *tag = NULL, *item = NULL, *value = NULL;
+    int argc;
+    struct template *t = NULL;
+    struct question *q = NULL;
+    char *out;
+
+    argc = strcmdsplit(arg, argv, DIM(argv));
+    CHECKARGC(== 3);
+
+    tag = argv[0];
+    item = argv[1];
+
+    value = unescapestr(argv[2]);
+
+    t = mod->templates->methods.get(mod->templates, tag);
+    if (t == NULL)
+    {
+        t = template_new(tag);
+        mod->templates->methods.set(mod->templates, t);
+        q = mod->questions->methods.get(mod->questions, t->tag);
+        if (q == NULL) {
+            q = question_new(t->tag);
+            q->template = t;
+            template_ref(t);
+        }
+        template_lset(t, NULL, item, value);
+        mod->questions->methods.set(mod->questions, q);
+        question_deref(q);
+    }
+    else
+        template_lset(t, NULL, item, value);
+    template_deref(t);
+
+    asprintf(&out, "%u OK", CMDSTATUS_SUCCESS);
+    return out;
+}
+
+/* vim: expandtab sw=4
+*/
