@@ -14,6 +14,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <locale.h>
+#include <stdbool.h>
 
 
 #include "confmodule.h"
@@ -36,7 +37,7 @@ struct option g_dpc_args[] = {
 	{ "default-priority", 1, NULL, 'd' },
 	{ "all", 0, NULL, 'a' },
 	{ "unseen-only", 0, NULL, 'u' },
-    { "force", 0, NULL, 'F' },
+	{ "force", 0, NULL, 'F' },
 	{ 0, 0, 0, 0 }
 };
 
@@ -57,16 +58,16 @@ static void cleanup(void)
 {
 	if (g_frontend != NULL)
 		frontend_delete(g_frontend);
-    if (g_templates != NULL)
-    {
-        g_templates->methods.save(g_templates);
-        template_db_delete(g_templates);
-    }
-    if (g_questions != NULL)
-    {
-        g_questions->methods.save(g_questions);
-        question_db_delete(g_questions);
-    }
+	if (g_templates != NULL)
+	{
+		g_templates->methods.save(g_templates);
+		template_db_delete(g_templates);
+	}
+	if (g_questions != NULL)
+	{
+		g_questions->methods.save(g_questions);
+		question_db_delete(g_questions);
+	}
 	if (g_config != NULL)
 		config_delete(g_config);
 }
@@ -101,17 +102,17 @@ static void usage(void)
  * Function: file_exists
  * Inputs: filename - name of file to check
  *         mode - minimum mode requirement (mask)
- * Outputs: 1 if filename exists and is a regular file, 0 otherwise
+ * Outputs: true if filename exists and is a regular file, false otherwise
  * Description: checks to see if a file exists
  * Assumptions: none
  ************************************************************************/
-static int file_exists(const char *filename, mode_t mode)
+static bool file_exists(const char *filename, mode_t mode)
 {
 	struct stat buf;
 	stat(filename, &buf);
-	if (!S_ISREG(buf.st_mode)) return 0;
-	if ((buf.st_mode & mode) == 0) return 0;
-	return 1;
+	if (!S_ISREG(buf.st_mode)) return false;
+	if ((buf.st_mode & mode) == 0) return false;
+	return true;
 }
 
 /************************************************************************
@@ -138,7 +139,13 @@ void loadtemplate(const char *filename, const char *owner)
 		{
 			q = question_new(t->tag);
 			q->template = t;
-                        template_ref(t);
+			template_ref(t);
+		}
+		else if (q->template != t)
+		{
+			template_deref(q->template);
+			q->template = t;
+			template_ref(t);
 		}
 		question_owner_add(q, owner);
 		if (g_questions->methods.set(g_questions, q) != DC_OK)
@@ -246,6 +253,7 @@ int runconfmodule(int argc, char **argv)
 	confmodule = confmodule_new(g_config, g_templates, g_questions, g_frontend);
 	confmodule->run(confmodule, argc, argv);
 	confmodule->communicate(confmodule);
+	confmodule->shutdown(confmodule);
 	ret = confmodule->exitcode;
 	confmodule_delete(confmodule);
 
@@ -276,14 +284,14 @@ int reconfigure(char **pkgs, int i, int max)
 		{
 			INFO(INFO_WARN, "%s is not installed, or does not use debconf", pkg);
 
-                        /* Don't die, though this doesn't have a
-                        config script.. we might be doing stuff with
-                        debian-installer, which doesn't use .config
-                        scripts.  Don't uncomment this without talking
-                        to -boot first
-                        -- tfheen, 2002-09-17 */
+			/* Don't die, though this doesn't have a
+			   config script.. we might be doing stuff with
+			   debian-installer, which doesn't use .config
+			   scripts.  Don't uncomment this without talking
+			   to -boot first
+			   -- tfheen, 2002-09-17 */
 
-                        /* continue; */
+			/* continue; */
 		}
 		/* startup the confmodule; run the config script and talk to it */
 		g_frontend->methods.set_title(g_frontend, pkg);
@@ -303,45 +311,45 @@ int reconfigure(char **pkgs, int i, int max)
 			if ((ret = runconfmodule(4, argv)) != DC_OK)
 				return ret;
 		}
-        else
-        {
-            snprintf(filename, sizeof(filename), INFODIR "/%s.postinst", pkg);
-            if (file_exists(filename,S_IXUSR|S_IXGRP|S_IXOTH))
-            {
-                if (is_confmodule(filename))
-                {
-                    argv[1] = filename;
-                    argv[2] = "configure";
-                    argv[3] = getfield(pkg, VERSIONFIELD);
-                    if ((ret = runconfmodule(4, argv)) != 0)
-                        return ret;
-                }
-                else
-                {
-                    /* according to debconf:
-                     * Since postinst might run other programs that
-                     * use debconf, checkpoint the db state and
-                     * reinitialize when the script finishes 
-                     */
-                    g_templates->methods.save(g_templates);
-                    g_questions->methods.save(g_questions);
+		else
+		{
+			snprintf(filename, sizeof(filename), INFODIR "/%s.postinst", pkg);
+			if (file_exists(filename,S_IXUSR|S_IXGRP|S_IXOTH))
+			{
+				if (is_confmodule(filename))
+				{
+					argv[1] = filename;
+					argv[2] = "configure";
+					argv[3] = getfield(pkg, VERSIONFIELD);
+					if ((ret = runconfmodule(4, argv)) != 0)
+						return ret;
+				}
+				else
+				{
+					/* according to debconf:
+					 * Since postinst might run other programs that
+					 * use debconf, checkpoint the db state and
+					 * reinitialize when the script finishes 
+					 */
+					g_templates->methods.save(g_templates);
+					g_questions->methods.save(g_questions);
+	
+					unsetenv("DEBIAN_HAS_FRONTEND");
+					setenv("DEBCONF_SHOWOLD", "1", 1);
 
-                    unsetenv("DEBIAN_HAS_FRONTEND");
-                    setenv("DEBCONF_SHOWOLD", "1", 1);
-
-                    strvacat(filename, sizeof(filename), " configure ", getfield(pkg, VERSIONFIELD), NULL);
-
-                    ret = system(filename);
-                    if (ret != 0) return DC_NOTOK;
-
-                    setenv("DEBIAN_HAS_FRONTEND", "1", 1);
-                    unsetenv("DEBCONF_SHOWOLD");
-
-                    g_templates->methods.load(g_templates);
-                    g_questions->methods.load(g_questions);
-                }
-            }
-        }
+					strvacat(filename, sizeof(filename), " configure ", getfield(pkg, VERSIONFIELD), NULL);
+	
+					ret = system(filename);
+					if (ret != 0) return DC_NOTOK;
+	
+					setenv("DEBIAN_HAS_FRONTEND", "1", 1);
+					unsetenv("DEBCONF_SHOWOLD");
+	
+					g_templates->methods.load(g_templates);
+					g_questions->methods.load(g_questions);
+				}
+			}
+		}
 
 		i++;
 	}
@@ -362,22 +370,22 @@ int main(int argc, char **argv)
 	signal(SIGINT, sighandler);
 	setlocale (LC_ALL, "");
 
-    if (getuid() != 0)
-        DIE("%s must be run as root", argv[0]);
+	if (getuid() != 0)
+		DIE("%s must be run as root", argv[0]);
 
 	g_config = config_new();
 
-	while ((opt = getopt_long(argc, argv, "dhf:p:aF", g_dpc_args, NULL) > 0))
+	while ((opt = getopt_long(argc, argv, "dhf:p:aF", g_dpc_args, NULL)) > 0)
 	{
 		switch (opt)
 		{
 		case 'h': usage(); break;
 		case 'f': setenv("DEBIAN_FRONTEND", optarg, 1); break;
-        case 'p': g_config->set(g_config, "_cmdline::priority", optarg); break;
+		case 'p': g_config->set(g_config, "_cmdline::priority", optarg); break;
 		case 'd': break;
 		case 'u': break;
 		case 'a': opt_all = 1; break;
-        case 'F': opt_force = 1; break;
+		case 'F': opt_force = 1; break;
 		}
 	}
 	g_config = config_new();
@@ -389,7 +397,7 @@ int main(int argc, char **argv)
 	/* initialize database and frontend modules */
 	if ((g_templates = template_db_new(g_config, NULL)) == 0)
 		DIE("Cannot initialize DebConf templates database");
-    if ((g_questions = question_db_new(g_config, g_templates, NULL)) == 0)
+	if ((g_questions = question_db_new(g_config, g_templates, NULL)) == 0)
 		DIE("Cannot initialize DebConf config database");
 	if ((g_frontend = frontend_new(g_config, g_templates, g_questions)) == 0)
 		DIE("Cannot initialize DebConf frontend");
