@@ -5,6 +5,10 @@
 #ifndef _FRONTEND_H_
 #define _FRONTEND_H_
 
+#include <stdbool.h>
+
+#include "constants.h"
+
 #undef _
 #define _(x) (x)
 
@@ -15,27 +19,32 @@ struct question;
 struct frontend;
 
 #define DCF_CAPB_BACKUP		(1UL << 0)
+#define DCF_CAPB_PROGRESSCANCEL	(1UL << 1)
 
 struct frontend_module {
     int (*initialize)(struct frontend *, struct configuration *);
     int (*shutdown)(struct frontend *);
     unsigned long (*query_capability)(struct frontend *);
     void (*set_title)(struct frontend *, const char *title);
+    void (*info)(struct frontend *, struct question *);
     int (*add)(struct frontend *, struct question *q);
     int (*go)(struct frontend *);
     void (*clear)(struct frontend *);
-    int (*can_go_back)(struct frontend *, struct question *);
-    int (*can_go_forward)(struct frontend *, struct question *);
+    bool (*can_go_back)(struct frontend *, struct question *);
+    bool (*can_go_forward)(struct frontend *, struct question *);
+    bool (*can_cancel_progress)(struct frontend *);
 
     void (*progress_start)(struct frontend *fe, int min, int max, const char *title);
-    void (*progress_set) (struct frontend *fe, int val);
+    int (*progress_set) (struct frontend *fe, int val);
     /* You do not have to implement _step, it will call _set by default */
-    void (*progress_step)(struct frontend *fe, int step);
-    void (*progress_info)(struct frontend *fe, const char *info);
+    int (*progress_step)(struct frontend *fe, int step);
+    int (*progress_info)(struct frontend *fe, const char *info);
     void (*progress_stop)(struct frontend *fe);
 };
 
 struct frontend {
+    /* module name */
+    char *name;
 	/* module handle */
 	void *handle;
 	/* configuration data */
@@ -55,11 +64,14 @@ struct frontend {
 	int interactive;
 	char *capb;
 	char *title;
+    struct question *info;
 	char *progress_title;
     int progress_min, progress_max, progress_cur;
 	
 	/* methods */
     struct frontend_module methods;
+    /* path to plugins */
+    char *plugin_path;
 };
 
 struct frontend *frontend_new(struct configuration *, struct template_db *, struct question_db *);
