@@ -7,7 +7,7 @@
  *
  * Description: Newt UI for cdebconf
  *
- * $Id: newt.c 16586 2004-06-02 03:54:20Z joshk $
+ * $Id: newt.c 49677 2007-10-02 17:12:21Z lunar $
  *
  * cdebconf is (c) 2000-2001 Randolph Chung and others under the following
  * license.
@@ -41,6 +41,7 @@
 #include "question.h"
 #include "frontend.h"
 #include "database.h"
+#include "plugin.h"
 #include "strutl.h"
 
 #include <ctype.h>
@@ -56,6 +57,7 @@
 #include <wchar.h>
 #include <assert.h>
 
+#include <slang.h>
 #include <newt.h>
 
 #ifdef HAVE_LIBTEXTWRAP
@@ -66,66 +68,97 @@ struct newt_data {
     newtComponent scale_form,
                   scale_bar,
                   scale_textbox,
+                  scale_cancel,
                   perc_label;
     int           scale_textbox_height;
+    char *        scale_info;
 };
 
-#define q_get_extended_description(q)   question_get_field((q), "", "extended_description")
-#define q_get_description(q)  		question_get_field((q), "", "description")
-#define q_get_choices(q)		question_get_field((q), "", "choices")
-#define q_get_choices_vals(q)		question_get_field((q), NULL, "choices")
-#define q_get_indices(q)		question_get_field((q), "", "indices")
+struct newtColors newtAltColorPalette = {
+	"white", "blue", 			/* root fg, bg */
+	/*"black", "lightgray",*/			/* border fg, bg */
+	"white", "black",			/* border fg, bg */
+	/*"black", "lightgray",*/			/* window fg, bg */
+	"white", "black",			/* window fg, bg */
+	/*"white", "black",*/			/* shadow fg, bg */
+	"white", "blue",			/* shadow fg, bg */
+	/*"red", "lightgray",*/			/* title fg, bg */
+	"yellow", "black",			/* title fg, bg */
+	"brightred", "gray",			/* button fg, bg */
+	"gray", "brightred",			/* active button fg, bg */
+	"white", "gray",			/* checkbox fg, bg */
+	"yellow", "brown",			/* active checkbox fg, bg */
+	"yellow", "blue",			/* entry box fg, bg */
+	/*"blue", "lightgray",*/			/* label fg, bg */
+	"brightred", "black",			/* label fg, bg */
+	/*"black", "lightgray",*/			/* listbox fg, bg */
+	"lightgray", "black",			/* listbox fg, bg */
+	"yellow", "blue",			/* active listbox fg, bg */
+	/*"black", "lightgray",*/			/* textbox fg, bg */
+	"white", "black",			/* textbox fg, bg */
+	"lightgray", "black",			/* active textbox fg, bg */
+	/*"white", "blue",*/			/* help line */
+	"white", "black",			/* help line */
+	"yellow", "blue",			/* root text */
+	"blue",					/* scale full */
+	"red",					/* scale empty */
+	"blue", "lightgray",			/* disabled entry fg, bg */
+	/*"black", "lightgray",*/			/* compact button fg, bg */
+	"lightgray", "black",			/* compact button fg, bg */
+	"yellow", "gray",			/* active & sel listbox */
+	"black", "brown"			/* selected listbox */
+};
 
-#define create_form(scrollbar)          newtForm((scrollbar), NULL, 0)
+typedef int (newt_handler)(struct frontend *obj, struct question *q);
 
-/*  Horizontal offset between buttons and text box */
-#define TEXT_PADDING 1
-/*  Horizontal offset between text box and borders */
-#define BUTTON_PADDING 4
+static void newt_progress_stop(struct frontend *obj);
 
-/*
- * Function: get_text
- * Input: struct frontend *obj - frontend object
- *        const char *template - template name
- *        const char *fallback - string to use if not available
- * Output: const char * - ptr to string, translated if possible
- * Description: get the translated version of a string
- * Assumptions: None.
+#include "cdebconf_newt.h"
+
+/*  Padding of title width, allows for leading "[!!] " before title
+ *  and the two vertical bars and spaces that limit the title.
  */
-static const char *
-get_text(struct frontend *obj, const char *template, const char *fallback )
-{
-	struct question *q = obj->qdb->methods.get(obj->qdb, template);
-	return q ? q_get_description(q) : fallback;
-}
+#define TITLE_PADDING 9
 
 /* gettext would be much nicer :-( */
 static const char *
 continue_text(struct frontend *obj)
 {
-    return get_text(obj, "debconf/button-continue", "Continue");
+    return question_get_text(obj, "debconf/button-continue", "Continue");
 }
 
 static const char *
 goback_text(struct frontend *obj)
 {
-    return get_text(obj, "debconf/button-goback", "Go Back");
+    return question_get_text(obj, "debconf/button-goback", "Go Back");
 }
 
 static const char *
 yes_text(struct frontend *obj)
 {
-    return get_text(obj, "debconf/button-yes", "Yes");
+    return question_get_text(obj, "debconf/button-yes", "Yes");
 }
 
 static const char *
 no_text(struct frontend *obj)
 {
-    return get_text(obj, "debconf/button-no", "No");
+    return question_get_text(obj, "debconf/button-no", "No");
+}
+ 
+static const char *
+cancel_text(struct frontend *obj)
+{
+    return question_get_text(obj, "debconf/button-cancel", "Cancel");
 }
 
-static void
-create_window(const int width, const int height, const char *title, const char *priority)
+static const char *
+help_text(struct frontend *obj)
+{
+    return question_get_text(obj, "debconf/help-line", "<Tab> moves between items; <Space> selects; <Enter> activates buttons");
+}
+
+void
+cdebconf_newt_create_window(const int width, const int height, const char *title, const char *priority)
 {
     static const char *sigils[][2] = {
         { "low",      "." },
@@ -154,8 +187,8 @@ create_window(const int width, const int height, const char *title, const char *
     }
 }
 
-static int
-get_text_height(const char *text, int win_width)
+int
+cdebconf_newt_get_text_height(const char *text, int win_width)
 {
     newtComponent textbox, f;
     int t_height;
@@ -183,14 +216,14 @@ get_text_height(const char *text, int win_width)
 #endif
     t_height = newtTextboxGetNumLines(textbox);
     // This is needed so the textbox gets freed...ick
-    f = create_form(NULL);
+    f = cdebconf_newt_create_form(NULL);
     newtFormAddComponent(f, textbox);
     newtFormDestroy(f);
     return t_height;
 }
 
-static int
-get_text_width(const char *text)
+int
+cdebconf_newt_get_text_width(const char *text)
 {
     int t_width = 0;
     const char *p = text;
@@ -220,7 +253,7 @@ min_window_height(struct question *q, int win_width)
 
     q_ext_text = q_get_extended_description(q);
     if (q_ext_text != NULL)
-        height = get_text_height(q_ext_text, win_width) + 1;
+        height = cdebconf_newt_get_text_height(q_ext_text, win_width) + 1;
     if (strcmp(type, "multiselect") == 0 || strcmp(type, "select") == 0)
         height += 4; // at least three lines for choices + blank line
     else if (strcmp(type, "string") == 0 || strcmp(type, "password") == 0)
@@ -298,7 +331,7 @@ show_separate_window(struct frontend *obj, struct question *q)
     win_width = width-7;
     /*  There are 5 characters for sigils, plus 4 for borders */
     strtruncate(obj->title, win_width-9);
-    t_height = get_text_height(full_description, win_width);
+    t_height = cdebconf_newt_get_text_height(full_description, win_width);
     if (t_height+extra <= height-5)
         win_height = t_height+extra;
     else {
@@ -314,23 +347,23 @@ show_separate_window(struct frontend *obj, struct question *q)
     free(full_description);
     full_description = wrappedtext;
 #endif
-    t_width = get_text_width(full_description);
-    t_width_descr = get_text_width(descr);
+    t_width = cdebconf_newt_get_text_width(full_description);
+    t_width_descr = cdebconf_newt_get_text_width(descr);
     if (t_width_descr > t_width)
         t_width = t_width_descr;
-    t_width_buttons = 2*BUTTON_PADDING + get_text_width(continue_text(obj)) + 2;
+    t_width_buttons = 2*BUTTON_PADDING + cdebconf_newt_get_text_width(continue_text(obj)) + 2;
     if (obj->methods.can_go_back(obj, q))
         //  Add an interspace
-        t_width_buttons += get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     if (win_width > t_width + 2*TEXT_PADDING + t_width_scroll)
         win_width = t_width + 2*TEXT_PADDING + t_width_scroll;
-    t_width_title = get_text_width(obj->title);
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
-    create_window(win_width, win_height, obj->title, q->priority);
-    form = create_form(NULL);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
+    form = cdebconf_newt_create_form(NULL);
     if (format_note)
         newtFormAddComponent(form, newtLabel((win_width - strwidth(descr))/2, 0, descr));
     textbox = newtTextbox(TEXT_PADDING, 1, t_width, t_height, flags);
@@ -364,7 +397,7 @@ show_separate_window(struct frontend *obj, struct question *q)
 static int
 generic_handler_string(struct frontend *obj, struct question *q, int eflags)
 {
-    newtComponent form, textbox, bCancel, entry, cRet;
+    newtComponent form, textbox, bOk, bCancel, entry, cRet;
     int width = 80, height = 24, t_height, t_width, win_width, win_height;
     int t_width_scroll = 0, t_width_title, t_width_buttons;
     int ret;
@@ -391,7 +424,7 @@ generic_handler_string(struct frontend *obj, struct question *q, int eflags)
     full_description = wrappedtext;
 #endif
     if (full_description != NULL)
-        t_height = get_text_height(full_description, win_width);
+        t_height = cdebconf_newt_get_text_height(full_description, win_width);
     else
         t_height = 0;
     if (t_height + 6 <= height-5)
@@ -402,20 +435,20 @@ generic_handler_string(struct frontend *obj, struct question *q, int eflags)
         t_width_scroll = 2;
     }
     t_height = win_height - 6;
-    t_width = get_text_width(full_description);
-    t_width_buttons = 2*BUTTON_PADDING;
+    t_width = cdebconf_newt_get_text_width(full_description);
+    t_width_buttons = 2*BUTTON_PADDING + cdebconf_newt_get_text_width(continue_text(obj)) + 2;
     if (obj->methods.can_go_back(obj, q))
         //  Add an interspace
-        t_width_buttons += get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     if (win_width > t_width + 2*TEXT_PADDING + t_width_scroll)
         win_width = t_width + 2*TEXT_PADDING + t_width_scroll;
-    t_width_title = get_text_width(obj->title);
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
-    create_window(win_width, win_height, obj->title, q->priority);
-    form = create_form(NULL);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
+    form = cdebconf_newt_create_form(NULL);
     textbox = newtTextbox(TEXT_PADDING, 1, t_width, t_height, tflags);
     assert(textbox);
     if (full_description != NULL)
@@ -423,17 +456,19 @@ generic_handler_string(struct frontend *obj, struct question *q, int eflags)
         newtTextboxSetText(textbox, full_description);
         free(full_description);
     }
-    if (eflags & NEWT_FLAG_HIDDEN || question_getvalue(q, "") == NULL)
+    if (eflags & NEWT_FLAG_PASSWORD || question_getvalue(q, "") == NULL)
         defval = "";
     else
         defval = (char *)question_getvalue(q, "");
     entry   = newtEntry(TEXT_PADDING, 1+t_height+1, defval, t_width, &result, eflags);
     if (obj->methods.can_go_back(obj, q)) {
+        bOk     = newtCompactButton(win_width - TEXT_PADDING - BUTTON_PADDING - strwidth(continue_text(obj)) - 3, win_height-2, continue_text(obj));
         bCancel = newtCompactButton(TEXT_PADDING + BUTTON_PADDING - 1, win_height-2, goback_text(obj));
-        newtFormAddComponents(form, bCancel, textbox, entry, NULL);
+        newtFormAddComponents(form, bCancel, textbox, entry, bOk, NULL);
     } else {
+        bOk     = newtCompactButton((win_width-strwidth(continue_text(obj))-2)/2 - 1, win_height-2, continue_text(obj));
         bCancel = NULL;
-        newtFormAddComponents(form, textbox, entry, NULL);
+        newtFormAddComponents(form, textbox, entry, bOk, NULL);
     }
     newtFormSetCurrent(form, entry);
     cRet = newtRunForm(form);
@@ -451,7 +486,7 @@ generic_handler_string(struct frontend *obj, struct question *q, int eflags)
 static int
 show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_desc)
 {
-    newtComponent form, sform = NULL, scrollbar, textbox, bOk, bCancel, cRet;
+    newtComponent form, sform = NULL, scrollbar, textbox, checkbox, bOk, bCancel, cRet;
     int width = 80, height = 24;
     int win_width, win_height = -1, t_height, t_width, sel_height, sel_width;
     int t_width_title, t_width_buttons;
@@ -487,7 +522,7 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     defcount = strchoicesplit(question_getvalue(q, ""), defvals, count);
     answer = malloc(sizeof(char) * count);
     sel_height = count;
-    form = create_form(NULL);
+    form = cdebconf_newt_create_form(NULL);
 #ifdef HAVE_LIBTEXTWRAP
     textwrap_init(&tw);
     textwrap_columns(&tw, win_width - 2 - 2*TEXT_PADDING);
@@ -496,11 +531,11 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     full_description = wrappedtext;
 #endif
     sel_width = strlongest(choices_trans, count);
-    t_width = get_text_width(full_description);
-    t_width_buttons = 2*BUTTON_PADDING + get_text_width(continue_text(obj)) + 2;
+    t_width = cdebconf_newt_get_text_width(full_description);
+    t_width_buttons = 2*BUTTON_PADDING + cdebconf_newt_get_text_width(continue_text(obj)) + 2;
     if (obj->methods.can_go_back(obj, q))
         //  Add an interspace
-        t_width_buttons += get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     //  Truncate too wide items
@@ -521,7 +556,7 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
         t_width = sel_width;
     if (win_width > t_width + 8)
         win_width = t_width + 8;
-    t_width_title = get_text_width(obj->title);
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
     if (show_ext_desc && full_description) {
@@ -539,7 +574,7 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
         win_height = height-5;
         sel_height = win_height - t_height - 5;
     }
-    create_window(win_width, win_height, obj->title, q->priority);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
     if (count > sel_height) {
         scrollbar = newtVerticalScrollbar((win_width+sel_width+5)/2, 1+t_height+1, sel_height,
                 NEWT_COLORSET_WINDOW, NEWT_COLORSET_ACTCHECKBOX);
@@ -547,7 +582,7 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     }
     else
         scrollbar = NULL;
-    sform = create_form(scrollbar);
+    sform = cdebconf_newt_create_form(scrollbar);
     newtFormSetBackground(sform, NEWT_COLORSET_CHECKBOX);
     newtFormSetHeight(sform, sel_height);
     newtFormSetWidth(sform, sel_width+5);
@@ -556,7 +591,9 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
         for (k = 0; k < defcount; k++)
             if (strcmp(choices[tindex[i]], defvals[k]) == 0)
                 def = 1;
-        newtFormAddComponent(sform, newtCheckbox((win_width-sel_width-3)/2, 1+t_height+1+i, choices_trans[i], def ? '*' : ' ', " *", &answer[tindex[i]]));
+        checkbox = newtCheckbox((win_width-sel_width-3)/2, 1+t_height+1+i, choices_trans[i], def ? '*' : ' ', " *", &answer[tindex[i]]);
+        newtCheckboxSetFlags(checkbox, NEWT_FLAG_RETURNEXIT, NEWT_FLAGS_SET);
+        newtFormAddComponent(sform, checkbox);
     }
     if (obj->methods.can_go_back(obj, q) || !show_ext_desc) {
         bOk     = newtCompactButton(win_width - TEXT_PADDING - BUTTON_PADDING - strwidth(continue_text(obj)) - 3, win_height-2, continue_text(obj));
@@ -638,7 +675,7 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
         return DC_NOTOK;
 
     sel_height = count;
-    form = create_form(NULL);
+    form = cdebconf_newt_create_form(NULL);
 #ifdef HAVE_LIBTEXTWRAP
     textwrap_init(&tw);
     textwrap_columns(&tw, win_width - 2 - 2*TEXT_PADDING);
@@ -647,11 +684,11 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     full_description = wrappedtext;
 #endif
     sel_width = strlongest(choices_trans, count);
-    t_width = get_text_width(full_description);
+    t_width = cdebconf_newt_get_text_width(full_description);
     t_width_buttons = 2*BUTTON_PADDING;
     if (obj->methods.can_go_back(obj, q))
         //  Add an interspace
-        t_width_buttons += get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     //  Truncate too wide items
@@ -672,7 +709,7 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
         t_width = sel_width;
     if (win_width > t_width + 8)
         win_width = t_width + 8;
-    t_width_title = get_text_width(obj->title);
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
     if (show_ext_desc && full_description) {
@@ -692,7 +729,7 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     }
     if (count > sel_height)
         listflags |= NEWT_FLAG_SCROLL;
-    create_window(win_width, win_height, obj->title, q->priority);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
     listbox = newtListbox((win_width-sel_width-3)/2, 1+t_height+1, sel_height, listflags);
     defval = (char *)question_getvalue(q, "");
     for (i = 0; i < count; i++) {
@@ -764,7 +801,7 @@ newt_handler_boolean(struct frontend *obj, struct question *q)
     full_description = wrappedtext;
 #endif
     if (full_description != NULL)
-        t_height = get_text_height(full_description, win_width);
+        t_height = cdebconf_newt_get_text_height(full_description, win_width);
     else
         t_height = 0;
     if (t_height + 4 <= height-5)
@@ -775,23 +812,23 @@ newt_handler_boolean(struct frontend *obj, struct question *q)
         t_width_scroll = 2;
     }
     t_height = win_height - 4;
-    t_width = get_text_width(full_description);
+    t_width = cdebconf_newt_get_text_width(full_description);
     t_width_buttons = 2*BUTTON_PADDING;
-    t_width_buttons += get_text_width(yes_text(obj)) + 3;
-    t_width_buttons += get_text_width(no_text(obj)) + 3;
+    t_width_buttons += cdebconf_newt_get_text_width(yes_text(obj)) + 3;
+    t_width_buttons += cdebconf_newt_get_text_width(no_text(obj)) + 3;
     if (obj->methods.can_go_back(obj, q))
         t_width_buttons += 3;
     if (obj->methods.can_go_back(obj, q))
-        t_width_buttons += get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     if (win_width > t_width + 2*TEXT_PADDING + t_width_scroll)
         win_width = t_width + 2*TEXT_PADDING + t_width_scroll;
-    t_width_title = get_text_width(obj->title);
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
-    create_window(win_width, win_height, obj->title, q->priority);
-    form = create_form(NULL);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
+    form = cdebconf_newt_create_form(NULL);
     textbox = newtTextbox(TEXT_PADDING, 1, t_width, t_height, flags);
     assert(textbox);
     if (full_description != NULL)
@@ -908,7 +945,7 @@ newt_handler_string(struct frontend *obj, struct question *q)
  */
 static int newt_handler_password(struct frontend *obj, struct question *q)
 {
-    return generic_handler_string(obj, q, NEWT_FLAG_HIDDEN);
+    return generic_handler_string(obj, q, NEWT_FLAG_PASSWORD);
 }
 
 /*
@@ -936,22 +973,32 @@ static int newt_handler_note(struct frontend *obj, struct question *q)
 static int
 newt_handler_text(struct frontend *obj, struct question *q)
 {
-    /* FIXME: Is really text an interactive question type? */
+    /* FIXME: Is text really an interactive question type? */
     return newt_handler_note(obj, q);
 }
 
 static int
 newt_handler_error(struct frontend *obj, struct question *q)
 {
-    char *oldcolor;
+    char *oldrootBg, *oldshadowBg, *oldhelpLineFg, *oldhelpLineBg;
     int ret;
-    struct newtColors palette = newtDefaultColorPalette;
+    struct newtColors palette = newtAltColorPalette;
 
-    oldcolor = palette.rootBg;
+    oldrootBg = palette.rootBg;
+    oldshadowBg = palette.shadowBg;
+    oldhelpLineFg = palette.helpLineFg;
+    oldhelpLineBg = palette.helpLineBg;
+    if (strcmp(oldrootBg, oldshadowBg) == 0)
+    	palette.shadowBg = "red";
     palette.rootBg = "red";
+    palette.helpLineFg = "white";
+    palette.helpLineBg = "red";
     newtSetColors(palette);
     ret = newt_handler_note(obj, q);
-    palette.rootBg = oldcolor;
+    palette.rootBg = oldrootBg;
+    palette.shadowBg = oldshadowBg;
+    palette.helpLineFg = oldhelpLineFg;
+    palette.helpLineBg = oldhelpLineBg;
     newtSetColors(palette);
     return ret;
 }
@@ -959,7 +1006,7 @@ newt_handler_error(struct frontend *obj, struct question *q)
 /* ----------------------------------------------------------------------- */
 struct question_handlers {
 	const char *type;
-	int (*handler)(struct frontend *obj, struct question *q);
+	newt_handler *handler;
 } question_handlers[] = {
 	{ "boolean",	newt_handler_boolean },         // OK
 	{ "multiselect", newt_handler_multiselect },
@@ -968,28 +1015,35 @@ struct question_handlers {
 	{ "password",	newt_handler_password },        // OK
 	{ "note",	newt_handler_note },            // OK
 	{ "text",	newt_handler_text },
-        { "error",      newt_handler_error },
+	{ "error",	newt_handler_error },
+	{ "",		NULL },
 };
 
 /*
- * Function: newt_intitialize
+ * Function: newt_initialize
  * Input: struct frontend *obj - frontend UI object
  *        struct configuration *cfg - configuration parameters
  * Output: int - DC_OK, DC_NOTOK
  * Description: initializes the text UI
  * Assumptions: none
  *
- * TODO: SIGINT is ignored by the newt UI, otherwise it interfers with
+ * TODO: SIGINT is ignored by the newt UI, otherwise it interferes with
  * the way question/answers are handled. this is probably not optimal
  */
 static int
 newt_initialize(struct frontend *obj, struct configuration *conf)
 {
     int i, width = 80, height = 24;
+    const char *palette;
 
     obj->interactive = 1;
     obj->data = calloc(1, sizeof(struct newt_data));
+    SLang_init_tty(0, 1, 0); /* disable flow control */
     newtInit();
+    palette = getenv("FRONTEND_BACKGROUND");
+    if (palette == NULL || strcmp(palette, "dark") != 0)
+        newtAltColorPalette = newtDefaultColorPalette;
+    newtSetColors(newtAltColorPalette);
     newtGetScreenSize(&width, &height);
     // Fill the screen so people can shift-pgup properly
     for (i = 0; i < height; i++)
@@ -998,12 +1052,6 @@ newt_initialize(struct frontend *obj, struct configuration *conf)
     return DC_OK;
 }
 
-static int
-newt_shutdown(struct frontend *obj)
-{
-    return DC_OK;
-}
-
 
 /*
  * Function: newt_go
@@ -1022,13 +1070,39 @@ newt_go(struct frontend *obj)
     cleared = 0;
     while (q != NULL) {
         for (i = 0; i < DIM(question_handlers); i++) {
-            if (strcmp(q->template->type, question_handlers[i].type) == 0) {
+            newt_handler *handler;
+            struct plugin *plugin = NULL;
+
+            if (*question_handlers[i].type)
+                handler = question_handlers[i].handler;
+            else {
+                plugin = plugin_find(obj, q->template->type);
+                if (plugin) {
+                    INFO(INFO_DEBUG, "Found plugin for %s", q->template->type);
+                    handler = (newt_handler *) plugin->handler;
+                } else {
+                    INFO(INFO_DEBUG, "No plugin for %s", q->template->type);
+                    continue;
+                }
+            }
+
+            if (plugin || strcmp(q->template->type, question_handlers[i].type) == 0) {
                 if (!cleared && !data->scale_form) {
                     cleared = 1;
+                    SLang_init_tty(0, 1, 0); /* disable flow control */
                     newtInit();
+                    newtSetColors(newtAltColorPalette);
                     newtCls();
                 }
-                ret = question_handlers[i].handler(obj, q);
+                if (obj->info != NULL) {
+                    char *text = q_get_description(obj->info);
+                    if (text)
+                        newtDrawRootText(0, 0, text);
+                    free(text);
+                }
+                newtPushHelpLine(help_text(obj));
+                ret = handler(obj, q);
+                newtPopHelpLine();
                 if (ret == DC_OK)
                     obj->qdb->methods.set(obj->qdb, q);
                 else if (ret == DC_GOBACK && q->prev != NULL)
@@ -1036,8 +1110,12 @@ newt_go(struct frontend *obj)
                 else {
                     if (cleared && !data->scale_form)
                         newtFinished();
+                    if (plugin)
+                        plugin_delete(plugin);
                     return ret;
                 }
+                if (plugin)
+                    plugin_delete(plugin);
                 break;
             }
         }
@@ -1049,48 +1127,131 @@ newt_go(struct frontend *obj)
     return DC_OK;
 }
 
-static int
+static bool
 newt_can_go_back(struct frontend *obj, struct question *q)
 {
     return (obj->capability & DCF_CAPB_BACKUP);
 }
 
+static bool
+newt_can_cancel_progress(struct frontend *obj)
+{
+    return (obj->capability & DCF_CAPB_PROGRESSCANCEL);
+}
+
+static void
+newt_make_progress_bar(struct frontend *obj, const char *info)
+{
+    struct newt_data *data = (struct newt_data *)obj->data;
+    int width = 80, height = 24, win_width, win_height, text_height;
+    int extra = 0;
+    bool can_cancel_progress;
+    bool changed = false;
+#ifdef HAVE_LIBTEXTWRAP
+    textwrap_t tw;
+    char *wrappedtext;
+    int flags = 0;
+#else
+    int flags = NEWT_FLAG_WRAP;
+#endif
+
+    can_cancel_progress = obj->methods.can_cancel_progress(obj);
+    if (can_cancel_progress)
+        extra += 2;
+    newtGetScreenSize(&width, &height);
+    win_width = width-7;
+    if (!info)
+        info = data->scale_info;
+    else
+        data->scale_info = strdup(info);
+    if (info)
+        text_height = cdebconf_newt_get_text_height(info, win_width);
+    else
+        text_height = 0;
+    /* Minimal height set to 2 to prevent box flashing */
+    if (text_height < 2)
+        text_height = 2;
+    if (data->scale_form) {
+        if (text_height != data->scale_textbox_height)
+            changed = true;
+        else if (can_cancel_progress && !data->scale_cancel)
+            changed = true;
+        else if (!can_cancel_progress && data->scale_cancel)
+            changed = true;
+        if (changed) {
+            newtFormDestroy(data->scale_form);
+            newtPopWindow();
+        }
+    }
+    if (!data->scale_form || changed) {
+        if (text_height + 3 + extra <= height - 5)
+            win_height = text_height + 3 + extra;
+        else
+            win_height = height - 5;
+        newtCenteredWindow(win_width, win_height, obj->progress_title);
+        data->scale_bar = newtScale(TEXT_PADDING, 1, win_width-2*TEXT_PADDING, obj->progress_max - obj->progress_min);
+        data->scale_textbox = newtTextbox(TEXT_PADDING, 3, win_width-2*TEXT_PADDING, text_height, flags);
+        data->scale_textbox_height = text_height;
+        data->scale_form = cdebconf_newt_create_form(NULL);
+        newtFormAddComponents(data->scale_form, data->scale_bar, data->scale_textbox, NULL);
+        if (obj->methods.can_cancel_progress(obj)) {
+            data->scale_cancel = newtCompactButton(TEXT_PADDING + BUTTON_PADDING - 1, win_height-2, cancel_text(obj));
+            newtFormAddComponent(data->scale_form, data->scale_cancel);
+        } else
+            data->scale_cancel = NULL;
+        newtFormSetTimer(data->scale_form, 1);
+    }
+    newtScaleSet(data->scale_bar, obj->progress_cur - obj->progress_min);
+    if (info) {
+#ifdef HAVE_LIBTEXTWRAP
+        textwrap_init(&tw);
+        textwrap_columns(&tw, win_width - 2 - 2*TEXT_PADDING);
+        wrappedtext = textwrap(&tw, info);
+        newtTextboxSetText(data->scale_textbox, wrappedtext);
+        free(wrappedtext);
+#else
+        newtTextboxSetText(data->scale_textbox, info);
+#endif
+    }
+}
+
 static void
 newt_progress_start(struct frontend *obj, int min, int max, const char *title)
 {
     struct newt_data *data = (struct newt_data *)obj->data;
-    int width = 80, win_width;
-#ifdef HAVE_LIBTEXTWRAP
-    int flags = 0;
-#else
-    int flags = NEWT_FLAG_WRAP;
-#endif
 
+    if (data->scale_form != NULL)
+        /* Nested progress bars are not currently supported. If you try to
+         * use one anyway, tear down the old progress bar first.
+         */
+        newt_progress_stop(obj);
     DELETE(obj->progress_title);
     obj->progress_title = strdup(title);
     obj->progress_min = min;
     obj->progress_max = max;
     obj->progress_cur = min;
+    data->scale_info = NULL;
+    SLang_init_tty(0, 1, 0); /* disable flow control */
     newtInit();
+    newtSetColors(newtAltColorPalette);
     newtCls();
-    newtGetScreenSize(&width, NULL);
-    win_width = width-7;
-    strtruncate(obj->progress_title, win_width-4);
-    newtCenteredWindow(win_width, 5, obj->progress_title);
-    data->scale_bar = newtScale(TEXT_PADDING, 1, win_width-2*TEXT_PADDING, obj->progress_max - obj->progress_min);
-    /*  Minimal height set to 2 to prevent box flashing */
-    data->scale_textbox = newtTextbox(TEXT_PADDING, 3, win_width-2*TEXT_PADDING, 2, flags);
-    data->scale_textbox_height = 2;
-    data->scale_form = create_form(NULL);
-    newtFormAddComponents(data->scale_form, data->scale_bar, data->scale_textbox, NULL);
+    if (obj->info != NULL) {
+        char *text = q_get_description(obj->info);
+        if (text)
+            newtDrawRootText(0, 0, text);
+        free(text);
+    }
+    newt_make_progress_bar(obj, NULL);
     newtDrawForm(data->scale_form);
     newtRefresh();
 }
 
-static void
+static int
 newt_progress_set(struct frontend *obj, int val)
 {
     struct newt_data *data = (struct newt_data *)obj->data;
+    struct newtExitStruct es;
+    int ret;
 
     if (data->scale_form != NULL)
     {
@@ -1105,55 +1266,44 @@ newt_progress_set(struct frontend *obj, int val)
 	    sprintf(buf, "%3d%%", (int)perc);
 	    newtLabelSetText(data->perc_label, buf);
 	} */
-	newtScaleSet(data->scale_bar, obj->progress_cur - obj->progress_min);
-	newtDrawForm(data->scale_form);
-	newtRefresh();
-    }
+	newt_make_progress_bar(obj, NULL);
+	newtFormRun(data->scale_form, &es);
+	if (es.reason == NEWT_EXIT_TIMER || data->scale_cancel == NULL)
+	    ret = DC_OK;
+	else if (es.reason == NEWT_EXIT_COMPONENT && es.u.co == data->scale_cancel)
+	    ret = DC_GOBACK;
+	else if (es.reason == NEWT_EXIT_HOTKEY && es.u.key == NEWT_KEY_F12)
+	    ret = DC_GOBACK;
+	else
+	    ret = DC_OK;
+    } else
+	ret = DC_OK;
+
+    return ret;
 }
 
-static void
+static int
 newt_progress_info(struct frontend *obj, const char *info)
 {
     struct newt_data *data = (struct newt_data *)obj->data;
+    struct newtExitStruct es;
+    int ret;
 
     if (data->scale_form != NULL) {
-	int width, win_width, text_height;
-#ifdef HAVE_LIBTEXTWRAP
-	textwrap_t tw;
-	char *wrappedtext;
-	int flags = 0;
-#else
-	int flags = NEWT_FLAG_WRAP;
-#endif
+	newt_make_progress_bar(obj, info);
+	newtFormRun(data->scale_form, &es);
+	if (es.reason == NEWT_EXIT_TIMER || data->scale_cancel == NULL)
+	    ret = DC_OK;
+	else if (es.reason == NEWT_EXIT_COMPONENT && es.u.co == data->scale_cancel)
+	    ret = DC_GOBACK;
+	else if (es.reason == NEWT_EXIT_HOTKEY && es.u.key == NEWT_KEY_F12)
+	    ret = DC_GOBACK;
+	else
+	    ret = DC_OK;
+    } else
+	ret = DC_OK;
 
-	newtGetScreenSize(&width, NULL);
-	win_width = width-7;
-	text_height = get_text_height(info, win_width);
-	if (text_height < 2)
-	    text_height = 2;
-	if (text_height != data->scale_textbox_height) {
-	    newtFormDestroy(data->scale_form);
-	    newtPopWindow();
-	    newtCenteredWindow(win_width, 4 + text_height, obj->progress_title);
-	    data->scale_bar = newtScale(TEXT_PADDING, 1, win_width-2*TEXT_PADDING, obj->progress_max);
-	    newtScaleSet(data->scale_bar, obj->progress_cur);
-	    data->scale_textbox = newtTextbox(TEXT_PADDING, 3, win_width-2*TEXT_PADDING, text_height, flags);
-	    data->scale_textbox_height = text_height;
-	    data->scale_form = create_form(NULL);
-	    newtFormAddComponents(data->scale_form, data->scale_bar, data->scale_textbox, NULL);
-	}
-#ifdef HAVE_LIBTEXTWRAP
-	textwrap_init(&tw);
-	textwrap_columns(&tw, win_width - 2 - 2*TEXT_PADDING);
-	wrappedtext = textwrap(&tw, info);
-	newtTextboxSetText(data->scale_textbox, wrappedtext);
-	free(wrappedtext);
-#else
-	newtTextboxSetText(data->scale_textbox, info);
-#endif
-	newtDrawForm(data->scale_form);
-	newtRefresh();
-    }
+    return ret;
 }
 
 static void
@@ -1165,16 +1315,27 @@ newt_progress_stop(struct frontend *obj)
         newtFormDestroy(data->scale_form);
         newtPopWindow();
         newtFinished();
-        data->scale_form = data->scale_bar = data->perc_label = data->scale_textbox = NULL;
+        data->scale_form = data->scale_bar = data->perc_label = data->scale_textbox = data->scale_cancel = NULL;
+        free(data->scale_info);
+        data->scale_info = NULL;
     }
 }
 
+static int
+newt_shutdown(struct frontend *obj)
+{
+    newt_progress_stop(obj);
+    newtFinished();
+    return DC_OK;
+}
+
 struct frontend_module debconf_frontend_module =
 {
 initialize: newt_initialize,
 shutdown: newt_shutdown,
 go: newt_go,
 can_go_back: newt_can_go_back,
+can_cancel_progress: newt_can_cancel_progress,
 progress_start: newt_progress_start,
 progress_set:   newt_progress_set,
 progress_info:  newt_progress_info,
